/*!
 * @header CA-Lava Grid I/O
 * Routines for loading and dumping the simulation grid
 */

#ifndef __CA_LAVA_GRIDIO__
#define __CA_LAVA_GRIDIO__

#include <stdlib.h>
#include <netcdf.h>

#include "ca-lava-data.h"
#include "v2int.h"
#include "nc_check.h"
#include "exn.h"
#include "utils.h"
#include "config.h"

/*!
 * @typedef gridio
 * Opaque descriptor structure to store information on an open simulation
 * grid.
 */
typedef struct gridio {
  int ncid;
  int t_id, z_id, h_id, T_id, /*v_id,*/ v_x_id, v_y_id, i_id, U_id, o_id;
  size_t start[4], count[4];
  ptrdiff_t imap[4];
} gridio;

/*!
 * @function gridio_open
 * Opens the given datafile. May throw exceptions of type EXN_CODE_NETCDF.
 * @param info opaque descriptor structure is filled with information on
 * the open dataset
 * @param dim dimensions of the grid are stored here
 * @param fn name of datafile to open
 * @return newly allocated gridio handle
 */
#ifndef INSTANTIATE
static inline
#endif
gridio *gridio_open(v2int *dim, const char *fn) {
  gridio *info;
  int i;

  if (!(info = NEW(struct gridio)))
    EXN_THROW(EXN_CODE_STDLIB, "couldn't allocate gridio descriptor");

  for (i = 0; i < 4; i++)
    info->start[i] = 0;

  info->count[0] = 1;
  info->count[1] = 0;
  info->count[2] = 0;
  info->count[3] = 2;

  NC_CHECK(nc_open(fn, NC_WRITE | NC_NOFILL, &info->ncid),
	   "couldn't open simulation grid");
  
  NC_CHECK(nc_inq_dimid(info->ncid, "x", &i),
	   "couldn't find dimension x");
  NC_CHECK(nc_inq_dimlen(info->ncid, i, &info->count[2]),
	   "couldn't read dimension x");
  dim->x = info->count[2];
  
  NC_CHECK(nc_inq_dimid(info->ncid, "y", &i),
	   "couldn't find dimension y");
  NC_CHECK(nc_inq_dimlen(info->ncid, i, &info->count[1]),
	   "couldn't read dimension y");
  dim->y = info->count[1];

  info->imap[0] = sizeof(cell) / sizeof(float) * dim->x * dim->y;
  info->imap[1] = sizeof(cell) / sizeof(float) * dim->x;
  info->imap[2] = sizeof(cell) / sizeof(float);
  info->imap[3] = 1;

  NC_CHECK(nc_inq_dimid(info->ncid, "t", &i),
	   "couldn't find dimension t");
  NC_CHECK(nc_inq_dimlen(info->ncid, i, &info->start[0]),
	   "couldn't read dimension t");
  info->start[0]--;

  NC_CHECK(nc_inq_varid(info->ncid, "t", &info->t_id),
	   "couldn't find variable t");
  NC_CHECK(nc_inq_varid(info->ncid, "z", &info->z_id),
	   "couldn't find variable z");
  NC_CHECK(nc_inq_varid(info->ncid, "h", &info->h_id),
	   "couldn't find variable h");
  NC_CHECK(nc_inq_varid(info->ncid, "T", &info->T_id),
	   "couldn't find variable T");
  /*
  NC_CHECK(nc_inq_varid(info->ncid, "v", &info->v_id),
	   "couldn't find variable v");
  */
  NC_CHECK(nc_inq_varid(info->ncid, "v_x", &info->v_x_id),
	   "couldn't find variable v_x");
  NC_CHECK(nc_inq_varid(info->ncid, "v_y", &info->v_y_id),
	   "couldn't find variable v_y");
  NC_CHECK(nc_inq_varid(info->ncid, "i", &info->i_id),
	   "couldn't find variable i");
  NC_CHECK(nc_inq_varid(info->ncid, "U", &info->U_id),
	   "couldn't find variable U");
  NC_CHECK(nc_inq_varid(info->ncid, "o", &info->o_id),
	   "couldn't find variable o");

  return info;
}

/*!
 * @function gridio_close
 * Close grid datafile and free descriptor.
 */
#ifndef INSTANTIATE
static inline
#endif
void gridio_close(gridio *info) {
  nc_close(info->ncid);
  free(info);
}

/*!
 * @function gridio_get_parameters
 * Retrieves simulation parameters from the simulation grid.
 * May throw exceptions of type EXN_CODE_NETCDF.
 * @param info gridio handle
 * @param solidus where to store the solidus temperature
 * @param coverage where to store the crustal coverage coefficients
 * @param viscosity where to store the viscosity coefficients
 * @param cooling where to store the cooling coefficient
 * @param time where to store the time of the last datafile entry
 * @param timestep where to store the timestep
 */
#ifndef INSTANTIATE
static inline
#endif
void gridio_get_parameters(gridio *info, float *cell_length, float *solidus, float coverage[], float viscosity[], float *cooling, float *time, float *timestep) {
  NC_CHECK(nc_get_att_float(info->ncid, NC_GLOBAL, "cell_length", cell_length),
	   "couldn't read cell length");
  NC_CHECK(nc_get_att_float(info->ncid, NC_GLOBAL, "solidus", solidus),
	   "couldn't read solidus temperature");
  NC_CHECK(nc_get_att_float(info->ncid, NC_GLOBAL, "coverage", coverage),
	   "couldn't read coverage parameters");
  NC_CHECK(nc_get_att_float(info->ncid, NC_GLOBAL, "viscosity", viscosity),
	   "couldn't read viscosity parameters");
  NC_CHECK(nc_get_att_float(info->ncid, NC_GLOBAL, "cooling", cooling),
	   "couldn't read cooling coefficient");

  NC_CHECK(nc_get_var1_float(info->ncid, info->t_id, info->start, time),
	   "couldn't read t value");
  NC_CHECK(nc_get_att_float(info->ncid, info->t_id, "increment",
			    timestep),
	   "couldn't read increment attribute of t");
}


/*!
 * @function gridio_load
 * Reads the whole grid for a given timestep.
 * May throw exceptions of type EXN_CODE_NETCDF.
 * @param info gridio handle
 * @param board cell array that receives the data from the last step
 */
#ifndef INSTANTIATE
static inline
#endif
void gridio_load(gridio *info, cell *board) {
  NC_CHECK(nc_get_varm_float(info->ncid, info->z_id,
			     info->start, info->count, NULL, info->imap,
			     &board->z),
	   "couldn't read z values");
  NC_CHECK(nc_get_varm_float(info->ncid, info->h_id,
			     info->start, info->count, NULL, info->imap,
			     &board->h),
	   "couldn't read h values");
  NC_CHECK(nc_get_varm_float(info->ncid, info->T_id,
			     info->start, info->count, NULL, info->imap,
			     &board->T),
	   "couldn't read T values");
  /*
  NC_CHECK(nc_get_varm_float(info->ncid, info->v_id,
			     info->start, info->count, NULL, info->imap,
			     &board->v.a[0]),
	   "couldn't read v values");
  */
  NC_CHECK(nc_get_varm_float(info->ncid, info->v_x_id,
			     info->start, info->count, NULL, info->imap,
			     &board->v.x),
	   "couldn't read v_x values");
  NC_CHECK(nc_get_varm_float(info->ncid, info->v_y_id,
			     info->start, info->count, NULL, info->imap,
			     &board->v.y),
	   "couldn't read v_y values");

#ifdef FULL_GRID
  NC_CHECK(nc_get_varm_float(info->ncid, info->i_id,
			     info->start, info->count, NULL, info->imap,
			     &board->i),
	   "couldn't read i values");
  NC_CHECK(nc_get_varm_float(info->ncid, info->U_id,
			     info->start, info->count, NULL, info->imap,
			     &board->U),
	   "couldn't read U values");
  NC_CHECK(nc_get_varm_float(info->ncid, info->o_id,
			     info->start, info->count, NULL, info->imap,
			     &board->tmp_o.a[0]),
	   "couldn't read o values");
#else
  NC_CHECK(nc_get_varm_float(info->ncid, info->i_id,
			     &info->start[1], &info->count[1], NULL, &info->imap[1],
			     &board->i),
	   "couldn't read i values");
  NC_CHECK(nc_get_varm_float(info->ncid, info->U_id,
			     &info->start[1], &info->count[1], NULL, &info->imap[1],
			     &board->U),
	   "couldn't read U values");
  NC_CHECK(nc_get_varm_float(info->ncid, info->o_id,
			     &info->start[1], &info->count[1], NULL, &info->imap[1],
			     &board->tmp_o.a[0]),
	   "couldn't read o values");
#endif /* FULL_GRID */
}

/*!
 * @function gridio_append
 * Dumps the whole grid for the next timestep.
 * May throw exceptions of type EXN_CODE_NETCDF.
 * @param info gridio descriptor
 * @param time point in time the step that is to be written represents
 * @param board cell array with data to store
 */
#ifndef INSTANTIATE
static inline
#endif
void gridio_append(gridio *info, const float time, const cell *board) {
  info->start[0]++;

  NC_CHECK(nc_put_var1_float(info->ncid, info->t_id, info->start, &time),
	   "couldn't write t value");
  
  NC_CHECK(nc_put_varm_float(info->ncid, info->z_id,
			     info->start, info->count, NULL, info->imap,
			     &board->z),
	   "couldn't write z values");
  NC_CHECK(nc_put_varm_float(info->ncid, info->h_id,
			     info->start, info->count, NULL, info->imap,
			     &board->h),
	   "couldn't write h values");
  NC_CHECK(nc_put_varm_float(info->ncid, info->T_id,
			     info->start, info->count, NULL, info->imap,
			     &board->T),
	   "couldn't write T values");
  /*
  NC_CHECK(nc_put_varm_float(info->ncid, info->v_id,
			     info->start, info->count, NULL, info->imap,
			     &board->v.a[0]),
	   "couldn't write v values");
  */
  NC_CHECK(nc_put_varm_float(info->ncid, info->v_x_id,
			     info->start, info->count, NULL, info->imap,
			     &board->v.x),
	   "couldn't write v_x values");
  NC_CHECK(nc_put_varm_float(info->ncid, info->v_y_id,
			     info->start, info->count, NULL, info->imap,
			     &board->v.y),
	   "couldn't write v_y values");

#ifdef FULL_GRID
  NC_CHECK(nc_put_varm_float(info->ncid, info->i_id,
			     info->start, info->count, NULL, info->imap,
			     &board->i),
	   "couldn't write i values");
  NC_CHECK(nc_put_varm_float(info->ncid, info->U_id,
			     info->start, info->count, NULL, info->imap,
			     &board->U),
	   "couldn't write U values");
  NC_CHECK(nc_put_varm_float(info->ncid, info->o_id,
			     info->start, info->count, NULL, info->imap,
			     &board->tmp_o.a[0]),
	   "couldn't write o values");
#else
  NC_CHECK(nc_put_varm_float(info->ncid, info->i_id,
			     &info->start[1], &info->count[1], NULL, &info->imap[1],
			     &board->i),
	   "couldn't write i values");
  NC_CHECK(nc_put_varm_float(info->ncid, info->U_id,
			     &info->start[1], &info->count[1], NULL, &info->imap[1],
			     &board->U),
	   "couldn't write U values");
  NC_CHECK(nc_put_varm_float(info->ncid, info->o_id,
			     &info->start[1], &info->count[1], NULL, &info->imap[1],
			     &board->tmp_o.a[0]),
	   "couldn't write o values");
#endif /* FULL_GRID */
  
  NC_CHECK(nc_sync(info->ncid), "couldn't flush simulation grid file");
}

#endif /* __CA_LAVA_GRIDIO__ */

/* vim:set shiftwidth=2 softtabstop=2: */
