/*!
 * @header CA-Lava Serial
 * Single threaded transformation application program for debugging
 * or single processor usage.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "v2int.h"
#include "ca-lava-transform.h"
#include "exn.h"
#include "ca-lava-gridio.h"
#include "ca-lava-data.h"
#include "utils.h"

/*!
 * @function main
 * Entry point for the serial simulation task. Takes 3 command line
 * arguments.
 */
int main(int arg_count, char *args[]) {
  gridio *sim_grid = NULL;
  cell *board = NULL;
  const cell **topo = NULL;
  int *worker_tasks = NULL;
  worker_info *worker_infos = NULL;
  int i;
  
  if (arg_count != 4) {
    fprintf(stderr,
	    "Usage: %s <sim_grid> <step_count> <step_interval>\n"
	    "\n"
	    "  sim_grid      = name of initial state simulation grid file\n"
	    "  step_count    = number of simulation steps to run\n"
	    "  step_interval = steps between dumps to simulation grid\n\n",
	    args[0]);
    return 1;
  }

  puts("CA-Lava Debug Transform started");

  EXN_DURING {
    v2int dim, pos;
    int step_count, step_interval;
    float time;
    
    /*
     * Read command line parameters, try to open the datafile
     * and read more configuration parameters.
     */
    step_count = atoi(args[2]);
    step_interval = atoi(args[3]);

    sim_grid = gridio_open(&dim, args[1]);
    printf("dim = (%d, %d)\n", dim.x, dim.y);
    
    gridio_get_parameters(sim_grid,
			  &cell_length,
			  &solidus, coverage, viscosity, &cooling,
			  &time, &timestep);

    printf("cell length = %g,\n"
	   "solidus = %g, coverage = (%g, %g), viscosity = (%g, %g), cooling = %g\n"
	   "last datafile entry has t = %g (+ %g)\n",
	   cell_length,
	   solidus, coverage[0], coverage[1], viscosity[0], viscosity[1], cooling,
	   time, timestep);

    /*
     * Allocate a board and read the data values from the simulation
     * grid.
     */	
    if (!(board = NEWARRAY(cell, dim.x * dim.y)))
      EXN_THROW(EXN_CODE_STDLIB, "couldn't allocate board");

    puts("loading board");

    gridio_load(sim_grid, board);

    /*
     * Create a board topology cache
     */
    if (!(topo = NEWARRAY(const cell *, 4 * dim.x * dim.y)))
      EXN_THROW(EXN_CODE_STDLIB, "couldn't allocate board topology cache");

    for (pos.y = 1; pos.y < dim.y - 1; pos.y++)
      for (pos.x = 1; pos.x < dim.x - 1; pos.x++) {
	int ofs = V2INT_OFS(pos, dim), ofs4 = 4 * ofs;
	topo[ofs4 + DIR_NORTH] = &board[ofs - dim.x];
	topo[ofs4 + DIR_EAST] = &board[ofs + 1];
	topo[ofs4 + DIR_SOUTH] = &board[ofs + dim.x];
	topo[ofs4 + DIR_WEST] = &board[ofs - 1];
      }
    
    puts("simulation begins");    

    /*
     * For every step, walk the board and apply the transformation rules
     */
    for (i = 1, time += timestep; i <= step_count; i++, time += timestep) {
      int j;

      /* lava mixing step */
      for (pos.y = 1; pos.y < dim.y - 1; pos.y++)
	for (pos.x = 1; pos.x < dim.x - 1; pos.x++) {
	  j = V2INT_OFS(pos, dim);
	  tf_mix(&board[j], &topo[4 * j]);
	}

      /* cooling step */
      for (pos.y = 1; pos.y < dim.y - 1; pos.y++)
	for (pos.x = 1; pos.x < dim.x - 1; pos.x++) {
	  j = V2INT_OFS(pos, dim);
	  tf_cool(&board[j], &topo[4 * j]);
	}

      /* outflow recalculation step */
      for (pos.y = 1; pos.y < dim.y - 1; pos.y++)
	for (pos.x = 1; pos.x < dim.x - 1; pos.x++) {
	  j = V2INT_OFS(pos, dim);
	  tf_flow(&board[j], &topo[4 * j]);
	}

      /* store temporary velocities */
      for (pos.y = 1; pos.y < dim.y - 1; pos.y++)
	for (pos.x = 1; pos.x < dim.x - 1; pos.x++) {
	  j = V2INT_OFS(pos, dim);
	  board[j].v = board[j].tmp_v;
	}

#ifndef NDEBUG
      /* bogus value check */
      for (pos.y = 1; pos.y < dim.y - 1; pos.y++)
	for (pos.x = 1; pos.x < dim.x - 1; pos.x++) {
	  cell *this = &board[V2INT_OFS(pos, dim)];

	  if (this->h > 0.f) {
	    if ((this->T < 0.f) || isnan(this->T) || isinf(this->T))
	      fprintf(stderr,
		      "%s:%d: warning: step %d, cell (%d, %d):"
		      " bogus temperature = %g\n",
		      __FILE__, __LINE__, i, pos.x, pos.y, this->T);
	  }
	  else if ((this->h < 0.f) || isnan(this->h) || isinf(this->h))
	    fprintf(stderr,
		    "%s:%d: warning: step %d, cell (%d, %d):"
		    " bogus height = %g\n",
		    __FILE__, __LINE__, i, pos.x, pos.y, this->h);

	  for (j = 0; j < 2; j++) {
	    if (isnan(this->tmp_o.a[j]) || isinf(this->tmp_o.a[j]))
	      fprintf(stderr,
		      "%s:%d: warning: step %d, cell (%d, %d):"
		      " bogus outflux[%d] = %g\n",
		      __FILE__, __LINE__, i, pos.x, pos.y, j, this->tmp_o.a[j]);
	  }
	}
#endif /* NDEBUG */

      /*
       * Write into the datafile every so many steps
       */
      if ((i % step_interval == 0) || (i == step_count)) {
	printf("dumping step %d (t = %g) to simulation grid\n", i, time);

	gridio_append(sim_grid, time, board);

	puts("data written");
      }
    }
    
    puts("simulation complete");
  }
  EXN_ALWAYS {
    if (sim_grid) gridio_close(sim_grid);
    if (board) free(board);
    if (topo) free(topo);
    if (worker_tasks) free(worker_tasks);
    if (worker_infos) free(worker_infos);
  }
  EXN_HANDLER {
    fprintf(stderr, "%s:%d: ", exn_current.file, exn_current.line);
    switch (exn_current.code) {
    case EXN_CODE_STDLIB:
      perror(exn_current.msg);
      break;
    default:
      fputs(exn_current.msg, stderr);
      putc('\n', stderr);
    }
  }
  EXN_END

  return 0;
}

/* vim:set shiftwidth=2 softtabstop=2: */
