/*!
 * @header CA-Lava Transform
 * The transformation for a simulation step of the lava flows.
 */

#ifndef __CA_LAVA_TRANSFORM__
#define __CA_LAVA_TRANSFORM__

#include <math.h>
#include <stdio.h>

#include "v2float.h"
#include "ca-lava-data.h"
#include "utils.h"
#include "config.h"

/*!
 * @var timestep
 * Duration of a timestep
 */
extern float timestep;

/*!
 * @var cell_length
 * Length of a simulation cell
 */
extern float cell_length;

/*!
 * @var solidus
 * Lava solidus temperature
 */
extern float solidus;

/*!
 * @var coverage
 * Lava crust coverage coefficients
 */
extern float coverage[2];

/*!
 * @var viscosity
 * Lava viscosity coefficients
 */
extern float viscosity[2];

/*!
 * @var cooling
 * Lava cooling coefficient, modified by estimated crust cover
 */
extern float cooling;

/*!
 * @define FILL
 * Filling value used for "empty" variables.
 */
#define FILL (-9999.f)

/*!
 * @define EPS_H
 * Rounding error margin for height
 */
#define EPS_H (1.e-5f)

/*!
 * @define GRAVITY
 * Gravitational acceleration of the earth.
 */
#define GRAVITY (9.81)


/*!
 * @define D_X
 * Estimate the derivative of a board parameter in x direction
 * @param next neighbouring cells
 * @param p parameter in the cell structure
 * @return the derivative
 */
#define D_X(next, p)							  \
  (((next)[DIR_EAST]->p - (next)[DIR_WEST]->p) / (2 * cell_length))

/*!
 * @define D_Y
 * Estimate the derivative of a board parameter in y direction
 * @param next neighbouring cells
 * @param p parameter in the cell structure
 * @return the derivative
 */
#define D_Y(next, p)							  \
  (((next)[DIR_NORTH]->p - (next)[DIR_SOUTH]->p) / (2 * cell_length))

/*!
 * @define SLOPE_X
 * Estimate the derivative of a board parameter in x direction as
 * a slope angle.
 * @param next neighbouring cells
 * @param p parameter in the cell structure
 * @return the slope in radians
 */
#define SLOPE_X(next, p)						  \
  (atan2f((next)[DIR_EAST]->p - (next)[DIR_WEST]->p, (2 * cell_length)))

/*!
 * @define SLOPE_Y
 * Estimate the derivative of a board parameter in y direction as
 * a slope angle
 * @param next neighbouring cells
 * @param p parameter in the cell structure
 * @return the slope in radians
 */
#define SLOPE_Y(next, p)						  \
  (atan2f((next)[DIR_NORTH]->p - (next)[DIR_SOUTH]->p, (2 * cell_length)))

/*!
 * @define GRAD
 * Estimates the gradient of a board parameter. Combines D_X and D_Y into
 * a vector.
 * @param next neighbouring cells
 * @param p parameter in the cell structure
 * @return the gradient as a v2float
 */
#define GRAD(next, p)							  \
  (V2FLOAT(D_X(next, p), D_Y(next, p)))

/*!
 * @define SLOPE
 * Estimates the gradient of a board parameter as two slope angles.
 * Combines SLOPE_X and SLOPE_Y into a vector.
 * @param next neighbouring cells
 * @param p parameter in the cell structure
 * @return the gradient slopes as a v2float
 */
#define SLOPE(next, p)							  \
  (V2FLOAT(SLOPE_X(next, p), SLOPE_Y(next, p)))

/*!
 * @define LAPLACIAN
 * Estimate the sum of second derivatives of a board parameter.
 * @param this cell in question
 * @param next neighbouring cells
 * @param p parameter in the cell structure
 */
#define LAPLACIAN(this, next, p)					  \
  ((((next)[DIR_EAST]->p - 2*(this)->p + (next)[DIR_WEST]->p) +		  \
    ((next)[DIR_NORTH]->p - 2*(this)->p + (next)[DIR_SOUTH]->p))	  \
   / (cell_length * cell_length))

/*!
 * @define VLAPLACIAN
 * Estimate the sum of second derivatives of each element in a vectorial
 * board parameter.
 * @param this cell in question
 * @param next neighbouring cells
 * @param p parameter in the cell structure
 */
#define VLAPLACIAN(this, next, p)					  \
  (V2FLOAT(LAPLACIAN(this, next, p.x), LAPLACIAN(this, next, p.y)))

/*
 * @function tf_mix
 * First board state tranformation step. Mixes lava that stays in the
 * cell with that coming in from sources and neighbours. Calculates the
 * mixture temperature and new thickness of lava layer after in- and
 * outfluxes have been applied.
 * @param this the cell being transformed
 * @param next array of neighbour cells
 */
#ifndef INSTANTIATE
static inline
#endif
void tf_mix(cell *this, const cell *next[]) {
  int j;

  /* this cell */
  this->tmp_T = this->h * this->T;
  this->tmp_T += this->i * this->U;
  this->h += this->i;

  /* neighbours */
  for (j = 0; j < 4; j++) {
    float i =  next[j]->tmp_o.a[(j + 1) % 2];
    if (j < 2) i = -i;

    if (i > 0.f) {
      this->tmp_T += next[j]->T * i;
      this->h += i;
    }
  }

  /* average */
  if (this->h > 0.f)
    this->tmp_T /= this->h;
  else
    this->tmp_T = FILL;

  /* outflux */
  for (j = 0; j < 2; j++)
    this->h -= fabsf(this->tmp_o.a[j]);

  /*
   * ensure that the height stays positive if it only becomes negative
   * due to rounding errors
   */
  if (this->h < 0.f && this->h > -EPS_H) {
    this->h = 0;
    this->v = V2FLOAT0;
  }
}

/*
 * @function tf_cool
 * Second board state transformation step. Calculates the effect of
 * radiative lava cooling. Solidifies the lava if necessary.
 * @param this the cell being transformed
 * @param next array of neighbour cells
 */
#ifndef INSTANTIATE
static inline
#endif
void tf_cool(cell *this, const cell *next[]) {
  if (this->h > 0.f) {
    /* temperature after radiative cooling */
    this->T =
      this->tmp_T
      / powf(1.f + this->tmp_T*this->tmp_T*this->tmp_T * cooling
		   * coverage[0] * expf(-coverage[1] * V2FLOAT_ABS(this->v))
	           / this->h,
	     1.5f);

    /* solidify the lava if below solidus and seal the vent */
    if (this->T <= solidus) {
      this->z += this->h;
      this->h = 0.f;
      this->v = V2FLOAT0;
      this->i = 0.f;
    }
  }
}

/*
 * @function tf_flow
 * Third board state transformation step. Applies the navier stokes
 * equations and determines the outfluxes.
 * @param this the cell being transformed
 * @param next array of neighbour cells
 */
#ifndef INSTANTIATE
static inline
#endif
void tf_flow(cell *this, const cell *next[]) {
  if (this->h > 0.f) {
    v2float s, c;
   
    /* calculate stress */
    this->tmp_v = VLAPLACIAN(this, next, v);
    V2FLOAT_SMUL(this->tmp_v, viscosity[0] * expf(-viscosity[1] * this->T));

    /* determine slope of terrain */
    c = SLOPE(next, z);
    s = V2FLOAT(-sin(c.x), -sin(c.y));
    c = V2FLOAT(cos(c.x), cos(c.y));

    /* apply downhill gravity acceleration */
    V2FLOAT_SMUL(s, GRAVITY);
    V2FLOAT_ADD(this->tmp_v, s);
    
    /* apply mean pressure gradient */
    V2FLOAT_SMUL(c, GRAVITY / 2);
    V2FLOAT_MUL(c, GRAD(next, h));
    V2FLOAT_SUB(this->tmp_v, c);

    /* scale by timestep and add up to new velocity */
    V2FLOAT_SMUL(this->tmp_v, timestep);
    V2FLOAT_ADD(this->tmp_v, this->v);
  }
  else
    this->tmp_v = V2FLOAT0;

  {
    float a;

    /* determine lava shift range */
    this->tmp_o = V2FLOAT(fabsf(this->tmp_v.x) * timestep, fabsf(this->tmp_v.y) * timestep);

    /* only as much as there is can flow out */
    if (this->tmp_o.x >= cell_length)
      this->tmp_o.x = cell_length;
    if (this->tmp_o.y >= cell_length)
      this->tmp_o.y = cell_length;

    /* outflow over the edge is distributed to the two adjacent cells */
    a = fabsf(this->tmp_o.x * this->tmp_o.y) / 2;
    V2FLOAT_MUL(this->tmp_o, V2FLOAT(cell_length, cell_length));
    V2FLOAT_SUB(this->tmp_o, V2FLOAT(a, a));

    /* scale correctly and fix signs */
    V2FLOAT_SMUL(this->tmp_o, this->h / (cell_length*cell_length));

    if (this->tmp_v.x < 0)
      this->tmp_o.x = -this->tmp_o.x;
    if (this->tmp_v.y < 0)
      this->tmp_o.y = -this->tmp_o.y;
  }
}

#endif /* __CA_LAVA_TRANSFORM__ */

/* vim:set shiftwidth=2 softtabstop=2: */
