/*!
 * @header CA-Lava Master
 * Master program for a distributed lava flow simulation by cellular
 * automata. Reads initial state from a NetCDF file and machine
 * configuration from the command line. Spawns worker tasks after
 * loading the initial state, sends data to the worker threads and
 * collects results from them which are written back to the
 * datafile. The master does not handle the synchronization of the
 * workers' simulation block borders, though, which is done in a
 * decentralized fashion by the workers themselves.
 */

#include <stdlib.h>
#include <stdio.h>
#include <pvm3.h>

#include "v2int.h"
#include "nc_check.h"
#include "exn.h"
#include "ca-lava-gridio.h"
#include "ca-lava-data.h"
#include "utils.h"

/*!
 * @function main
 * Entry point for the master task. Takes 5 command line argument.
 */
int main(int arg_count, char *args[]) {
  if (arg_count != 6) {
    fprintf(stderr,
	    "Usage: %s <sim_grid> <board_width> <board_height> <step_count> <step_interval>\n"
	    "\n"
	    "  sim_grid      = name of initial state simulation grid file\n"
	    "  board_width   = width of board in simulation blocks\n"
	    "  board_heigth  = height of board in simulation blocks\n"
	    "  step_count    = number of simulation steps to run\n"
	    "  step_interval = steps between dumps to simulation grid\n\n",
	    args[0]);
    return 1;
  }

  EXN_DURING {
    int tid;
    if ((tid = pvm_mytid()) < 0)
      EXN_THROW(EXN_CODE_PVM, "couldn't get my TID");

    printf("CA-Lava Master t%x started\n", tid);

    {
      gridio *sim_grid = NULL;
      cell *board = NULL;
      int *worker_tasks = NULL;
      worker_info *worker_infos = NULL;

      EXN_DURING {
	v2int board_dim, cell_dim, grid_dim, pos;
	int step_count, step_interval;
	int i;
	float cell_length, solidus, coverage[2], viscosity[2], cooling, time, timestep;
	int worker_count;
	
	/*
	 * Read command line parameters, try to open the datafile
	 * and read more configuration parameters.
	 */
	board_dim.x = atoi(args[2]);
	board_dim.y = atoi(args[3]);
	step_count = atoi(args[4]);
	step_interval = atoi(args[5]);

	sim_grid = gridio_open(&grid_dim, args[1]);

	cell_dim = grid_dim;
	V2INT_SUB(cell_dim, V2INT(2, 2));
	V2INT_MOD(cell_dim, board_dim);
	if ((cell_dim.x != 0) || (cell_dim.y != 0))
	  EXN_THROW(EXN_CODE_ANY, "grid dimensions cannot be tiled this way");
	cell_dim = grid_dim;
	V2INT_SUB(cell_dim, V2INT(2, 2));
	V2INT_DIV(cell_dim, board_dim);
	V2INT_ADD(cell_dim, V2INT(2, 2));

	printf("grid_dim = (%d, %d), board_dim = (%d, %d)"
	       " => cell_dim = (%d, %d)\n",
	       grid_dim.x, grid_dim.y, board_dim.x, board_dim.y,
	       cell_dim.x, cell_dim.y);
	
	gridio_get_parameters(sim_grid,
			      &cell_length,
			      &solidus, coverage, viscosity, &cooling,
			      &time, &timestep);

	printf("cell length = %g,\n"
	       "solidus = %g, coverage = (%g, %g), viscosity = (%g, %g), cooling = %g\n"
	       "last datafile entry has t = %g (+ %g)\n",
	       cell_length,
	       solidus, coverage[0], coverage[1], viscosity[0], viscosity[1], cooling,
	       time, timestep);

	/*
	 * Allocate a board and read the data values from the simulation
	 * grid.
	 */	
	if (!(board = NEWARRAY(cell, grid_dim.x * grid_dim.y)))
	  EXN_THROW(EXN_CODE_STDLIB, "couldn't allocate board");

	puts("loading board");

	gridio_load(sim_grid, board);
	
	/*
	 * Spawn board_dim.x * board_dim.y workers and calculate who
	 * processes which part of the board
	 */
	puts("spawning workers");

	worker_count = board_dim.x * board_dim.y;
	if (!(worker_tasks = NEWARRAY(int, worker_count)))
	  EXN_THROW(EXN_CODE_STDLIB, "couldn't allocate TID array");
	if ((pvm_catchout(stdout) < 0)
	    || (pvm_spawn("ca-lava-worker", NULL,
			  PvmTaskDefault, NULL,
			  worker_count, worker_tasks) != worker_count))
	  EXN_THROW(EXN_CODE_PVM, "couldn't spawn workers");

	if (!(worker_infos = NEWARRAY(worker_info, worker_count)))
	  EXN_THROW(EXN_CODE_STDLIB, "couldn't allocate worker positions");
	for (pos.y = 0; pos.y < board_dim.y; pos.y++) {
	  for (pos.x = 0; pos.x < board_dim.x; pos.x++) {
	    int ofs = V2INT_OFS(pos, board_dim);
	    worker_infos[ofs].org = cell_dim;
	    V2INT_SUB(worker_infos[ofs].org, V2INT(2, 2));
	    V2INT_MUL(worker_infos[ofs].org, pos);
	  }
	}
	for (pos.y = 0; pos.y < board_dim.y; pos.y++) {
	  for (pos.x = 0; pos.x < board_dim.x; pos.x++) {
	    int ofs = V2INT_OFS(pos, board_dim);
	    v2int next = pos;
	    /* north */
	    V2INT_ADD(next, V2INT(0, -1));
	    worker_infos[ofs].next[DIR_NORTH] =
	      (next.y >= 0 ? worker_tasks[V2INT_OFS(next, board_dim)] : 0);
	    /* east */
	    V2INT_ADD(next, V2INT(1, 1));
	    worker_infos[ofs].next[DIR_EAST] =
	      (next.x < board_dim.x ? worker_tasks[V2INT_OFS(next, board_dim)] : 0);
	    /* south */
	    V2INT_ADD(next, V2INT(-1, 1));
	    worker_infos[ofs].next[DIR_SOUTH] =
	      (next.y < board_dim.y ? worker_tasks[V2INT_OFS(next, board_dim)] : 0);
	    /* west */
	    V2INT_ADD(next, V2INT(-1, -1));
	    worker_infos[ofs].next[DIR_WEST] =
	      (next.x >= 0 ? worker_tasks[V2INT_OFS(next, board_dim)] : 0);
	  }
	}

	/*
	 * Tell every worker which part of the board to process
	 * and send out chunks of the board
	 */
	puts("seeding workers");

	for (i = 0; i < worker_count; i++)
	  send_config(worker_tasks[i], &worker_infos[i], cell_dim,
		      step_count, step_interval,
		      timestep, cell_length,
		      solidus, coverage, viscosity, cooling);
	
	for (pos.y = 0; pos.y < board_dim.y; pos.y++) {
	  for (pos.x = 0; pos.x < board_dim.x; pos.x++) {
	    int ofs = V2INT_OFS(pos, board_dim);
	    send_board(worker_tasks[ofs], board, V2INT0, grid_dim,
		       worker_infos[ofs].org, cell_dim);
	  }
	}

	puts("simulation begins");

	/*
	 * Wait for data from the workers and write it into the datafile
	 */
	for (i = 1, time += timestep; i <= step_count; i++, time += timestep)
	  if ((i % step_interval == 0) || (i == step_count)) {
	    int k;
	    
	    printf("collecting board data for generation %d (t = %g)\n",
		   i, time);

	    for (k = 0; k < worker_count; k++)
	      recv_board(worker_tasks[k], board, V2INT0, grid_dim);

	    puts("dumping to simulation grid");

	    gridio_append(sim_grid, time, board);

	    puts("data written");
	  }

	puts("simulation complete");
      }
      EXN_ALWAYS {
	if (sim_grid) gridio_close(sim_grid);
	if (board) free(board);
	if (worker_tasks) free(worker_tasks);
	if (worker_infos) free(worker_infos);
      }
      EXN_HANDLER {
	exn_rethrow();
      }
      EXN_END
    }
  }
  EXN_HANDLER {
    fprintf(stderr, "%s:%d: ", exn_current.file, exn_current.line);
    switch (exn_current.code) {
    case EXN_CODE_STDLIB:
      perror(exn_current.msg);
      break;
    case EXN_CODE_PVM:
      pvm_perror((char *)exn_current.msg);
      break;
    default:
      fputs(exn_current.msg, stderr);
      putc('\n', stderr);
    }
  }
  EXN_END

  return (pvm_exit() < 0 ? 1 : 0);
}

/* vim:set shiftwidth=2 softtabstop=2: */
