/*
 * coord-lookup.c
 * Routines for coordinate lookup
 */

#include <stdlib.h>
#include <string.h>
#include <netcdf.h>

#include "coord_lookup.h"
#include "nc_check.h"
#include "exn.h"
#include "utils.h"
#include "config.h"

/*!
 * @var coord_nearby_range
 * Latitude or longitude difference between gridpoints.
 */
static float coord_nearby_range = 0;

/*!
 * @function coord_nearby
 * Comparison function for bsearch to map geographical coordinates to
 * gridpoints.
 * @param key coordinate being looked up
 * @param ref coordinate to compare to
 * @return -1 if *key < *ref, +1 if *key > *ref + coord_nearby_range,
 * 0 otherwise
 */
static int coord_nearby(const float *key, const float *ref) {
  if (*key < *ref)
    return -1;
  if (*key > *ref + coord_nearby_range)
    return +1;
  else
    return 0;
}

/*!
 * @function adtof
 * Converts an ASCII representation of a geographical coordinate into
 * a floating point number. The coordinate has the format d[:m[:s]]{N|E|S|W}.
 * The parts d, m and s can all be floating point numbers.
 * @param str the ASCII representation
 * @return the float representation
 */
static inline float adtof(const char *str) {
  char *p;
  float f;

  f = strtof(str, &p);
  if (*p == ':') {
    f += strtof(p+1, &p) / 60;
    if (*p == ':')
      f += strtof(p+1, &p) / 3600;
  }

  switch (*p) {
  case 'N':
  case 'n':
  case 'E':
  case 'e':
    break;
  case 'S':
  case 's':
  case 'W':
  case 'w':
    f = -f;
    break;
  default:
    EXN_THROW(EXN_CODE_ANY, "invalid geographical hemisphere indicator");
  }

  return f;
}

size_t coord_lookup(int grid, const char *coord, size_t *dim, const char *key) {
  size_t idx = 0;
  int id;

  NC_CHECK(nc_inq_dimid(grid, coord, &id),
	   "coordinate dimension not found in simulation grid");
  NC_CHECK(nc_inq_dimlen(grid, id, dim),
	   "unable to read coordinate dimension length from simulation grid");
  if (*dim < 2)
    EXN_THROW(EXN_CODE_ANY, "coordinate dimension length is less than 2");
  
  if ((strlen(key) > 0) && strchr("NnEeSsWw", key[strlen(key) - 1])) {
    float *values = NULL;

    EXN_DURING {
      float v, *match;

      NC_CHECK(nc_inq_varid(grid, coord, &id),
	       "no coordinate variable in simulation grid");
      if (!(values = NEWARRAY(float, *dim)))
	EXN_THROW(EXN_CODE_STDLIB, "unable to allocate coordinate array");
      NC_CHECK(nc_get_var_float(grid, id, values),
	       "unable to read coordinate values");

      coord_nearby_range = values[1] - values[0];
      v = adtof(key);
      match = (float *)bsearch(&v, values, *dim, sizeof(float),
			       (int (*)(const void *, const void *))&coord_nearby);
      
      if (match)
	idx = match - values;
      else
	EXN_THROW(EXN_CODE_ANY, "coordinate value out of range");
    }
    EXN_ALWAYS {
      if (values) free(values);
    }
    EXN_HANDLER {
      exn_rethrow();
    }
    EXN_END
  }
  else
    idx = atoi(key);
    
  return idx;
}

/* vim:set shiftwidth=2 softtabstop=2: */
