/*!
 * @header Exceptions
 * Exception handling interface based on setjmp / longjmp
 */

#ifndef __EXN__
#define __EXN__

#include <setjmp.h>

#include "config.h"

/*
 * Don't use directly
 */
void _exn_setHandler(jmp_buf *handler);

/*
 * Don't use directly
 */
void _exn_swapHandler(jmp_buf *handler);
  
/*
 * Don't use directly
 */
void _exn_throw(int line, const char *file, int code, const char *msg)
  __attribute__((noreturn));

/*!
 * @define EXN_THROW
 * Throws a new exception with the given class and message.
 * @param code some integer identifying the exception
 * @param msg some persistent string describing the exception
 */
#define EXN_THROW(code, msg) _exn_throw(__LINE__, __FILE__, (code), (msg))

/*!
 * @function exn_rethrow
 * Rethrows the current exception to the next handler.
 */
void exn_rethrow(void) __attribute__((noreturn));

#if 0
/*!
 * @define EXN_DURING
 * Starts a block in which exceptions can occur.
 */
#define EXN_DURING
#endif

#define EXN_DURING							\
  {									\
    jmp_buf _exn_localHandler;						\
    int _exn_handlerActivated;						\
    if (!(_exn_handlerActivated = setjmp(_exn_localHandler))) {		\
      _exn_swapHandler(&_exn_localHandler);

#if 0
/*!
 * @define EXN_ALWAYS
 * Starts a block of cleanup actions for an EXN_DURING block. These
 * actions are taken whether an exception occurs or not.
 */
#define EXN_ALWAYS
#endif

#define EXN_ALWAYS						       \
  }								       \
  _exn_setHandler(&_exn_localHandler);                                 \
  {

#if 0
/*!
 * @define EXN_HANDLER
 * Starts a handler block for an EXN_DURING block.
 */
#define EXN_HANDLER
#endif

#define EXN_HANDLER						       \
  }								       \
  if (_exn_handlerActivated) {                                         \
    _exn_setHandler(&_exn_localHandler);

#if 0
/*!
 * @define EXN_END
 * Ends an EXN_DURING and associated blocks.
 */
#define EXN_END
#endif

#define EXN_END							       \
  }								       \
  _exn_setHandler(&_exn_localHandler);                                 \
  }

/*!
 * @typedef exn
 * Structure describing an exception.
 * @field line line number in source on which the exception originated
 * @field file source file name in which the exception originated
 * @field code some integer identifying the type of exception
 * @field msg human readable string describing the exception
 */
typedef struct exn {
  int line;
  const char *file;
  int code;
  const char *msg;
} exn;

/*!
 * @var exn_current
 * The buffer for the currently active exception.
 */
extern exn exn_current;

#endif /* __EXN__ */

/* vim:set shiftwidth=2 softtabstop=2: */
