/*!
 * @header Simulation Grid Loader
 * Load topographical GMT/COARDS grid data into a prepared simulation
 * grid.
 */

#include <stdlib.h>
#include <stdio.h>
#include <netcdf.h>

#include "nc_check.h"
#include "exn.h"
#include "utils.h"
#include "config.h"

/*!
 * @function main
 * Entry point for the simulation grid loader. Takes 3 command line
 * arguments.
 */
int main(int arg_count, char *args[]) {
  if (arg_count != 4) {
    fprintf(stderr, "Usage: %s <simgrid> <step> <topogrid>\n", args[0]);
    return 1;
  }

  {
    int sim_grid = 0, topo_grid = 0;
    float *values = NULL;
    EXN_DURING {
      int id;
      size_t dim_x, dim_y;
      size_t z_idx[] = { 0, 0, 0 }, z_count[] = { 1, 0, 0 };
      
      puts("Opening files");

      NC_CHECK(nc_open(args[1], NC_WRITE, &sim_grid),
	       "unable to open simulation grid for updating");
      z_idx[0] = atoi(args[2]);
      NC_CHECK(nc_open(args[3], 0, &topo_grid),
	       "unable to open topographic grid for reading");
      
      puts("Collecting information");

      NC_CHECK(nc_inq_dimid(topo_grid, "x", &id),
	       "no x dimension in topographic grid");
      NC_CHECK(nc_inq_dimlen(topo_grid, id, &dim_x),
	       "unable to read x dimension length in topographic grid");
      
      NC_CHECK(nc_inq_dimid(topo_grid, "y", &id),
	       "no y dimension in topographic grid");
      NC_CHECK(nc_inq_dimlen(topo_grid, id, &dim_y),
	       "unable to read y dimension length in topographic grid");
      
      NC_CHECK(nc_inq_dimid(sim_grid, "x", &id),
	       "no x dimension in simulation grid");
      NC_CHECK(nc_inq_dimlen(sim_grid, id, &z_count[2]),
	       "unable to read x dimension length in simulation grid");
      
      NC_CHECK(nc_inq_dimid(sim_grid, "y", &id),
	       "no y dimension in simulation grid");
      NC_CHECK(nc_inq_dimlen(sim_grid, id, &z_count[1]),
	       "unable to read y dimension length in simulation grid");
      
      if ((z_count[2] != dim_x) || (z_count[1] != dim_y))
	EXN_THROW(EXN_CODE_ANY,
		"different grid dimensions in topographic and simulation grid");
      
      puts("Copying x coordinates");

      if (!(values = NEWARRAY(float, dim_x)))
	EXN_THROW(EXN_CODE_STDLIB, "unable to allocate x coordinate array");
      NC_CHECK(nc_inq_varid(topo_grid, "x", &id),
	       "no x coordinate array in topographic grid");
      NC_CHECK(nc_get_var_float(topo_grid, id, values),
	       "unable to read x coordinate array from topographic grid");
      NC_CHECK(nc_inq_varid(sim_grid, "x", &id),
	       "no x coordinate array in simulation grid");
      NC_CHECK(nc_put_var_float(sim_grid, id, values),
	       "unable to write x coordinate array to simulation grid");
      free(values); values = NULL;
      
      puts("Copying y coordinates");

      if (!(values = NEWARRAY(float, dim_y)))
	EXN_THROW(EXN_CODE_STDLIB, "unable to allocate y coordinate array");
      NC_CHECK(nc_inq_varid(topo_grid, "y", &id),
	       "no y coordinate array in topographic grid");
      NC_CHECK(nc_get_var_float(topo_grid, id, values),
	       "unable to read y coordinate array from topographic grid");
      NC_CHECK(nc_inq_varid(sim_grid, "y", &id),
	       "no y coordinate array in simulation grid");
      NC_CHECK(nc_put_var_float(sim_grid, id, values),
	       "unable to write y coordinate array to simulation grid");
      free(values); values = NULL;
      
      puts("Copying z coordinates");

      if (!(values = NEWARRAY(float, dim_x*dim_y)))
	EXN_THROW(EXN_CODE_STDLIB, "unable to allocate z coordinate array");
      NC_CHECK(nc_inq_varid(topo_grid, "z", &id),
	       "no x coordinate array in topographic grid");
      NC_CHECK(nc_get_var_float(topo_grid, id, values),
	       "unable to read z coordinate array from topographic grid");
      NC_CHECK(nc_inq_varid(sim_grid, "z", &id),
	       "no z coordinate array in simulation grid");
      NC_CHECK(nc_put_vara_float(sim_grid, id, z_idx, z_count, values),
	       "unable to write z coordinate array to simulation grid");
      free(values); values = NULL;

      puts("done");
    }
    EXN_ALWAYS {
      if (sim_grid != 0) nc_close(sim_grid);
      if (topo_grid != 0) nc_close(topo_grid);
      if (values) free(values);
    }
    EXN_HANDLER {
      fprintf(stderr, "%s:%d: ", exn_current.file, exn_current.line);
      switch (exn_current.code) {
      case EXN_CODE_STDLIB:
	perror(exn_current.msg);
	break;
      default:
	fputs(exn_current.msg, stderr);
	putc('\n', stderr);
      }
    }
    EXN_END
  }

  return 0;
}

/* vim:set shiftwidth=2 softtabstop=2: */
