/*!
 * @header Simulation Grid Line Peek Utility
 * Fetches data values along a line from a prepared simulation grid.
 */

#include <stdlib.h>
#include <stdio.h>
#include <netcdf.h>

#include "coord_lookup.h"
#include "nc_check.h"
#include "exn.h"
#include "utils.h"
#include "config.h"

/*!
 * @function main
 * Entry point for the simulation grid loader. Takes 5 command line
 * arguments.
 */
int main(int arg_count, char *args[]) {
  if ((arg_count != 8) && (arg_count != 9)) {
    fprintf(stderr,
	    "Usage: %s <simgrid> <step> <var> <x0> <y0> <x1> <y1> [<value>]\n"
	    "\n"
	    "  If step == -1, the variable is assumed to be time independent.\n"
	    "\n"
	    "  x and y can either be integer indices or geographical\n"
	    "  longitude and latitude coordinates in the form\n"
	    "  d[:m[:s]]{N|E|S|W}, where d, m and s can all be floating\n"
	    "  point numbers.\n"
	    "\n"
	    "  If the optional value argument is given, it is a floating\n"
	    "  point value to be used as a replacement for the original\n"
	    "  value of the referenced cell.\n\n",
	    args[0]);
    return 1;
  }

  {
    int grid = 0;
    EXN_DURING {
      int id;
      size_t dim_x, dim_y;
      size_t x1, y1;
      size_t idx[] = { 0, 0, 0 };
      float v, nv;
      
      NC_CHECK(nc_open(args[1], (arg_count == 9 ? NC_WRITE : 0), &grid),
	       "unable to open simulation grid");

      idx[0] = atoi(args[2]);

      idx[2] = coord_lookup(grid, "x", &dim_x, args[4]);
      idx[1] = coord_lookup(grid, "y", &dim_y, args[5]);
      x1 = coord_lookup(grid, "x", &dim_x, args[6]);
      y1 = coord_lookup(grid, "y", &dim_y, args[7]);

      if (arg_count == 9)
	nv = atof(args[8]);

      NC_CHECK(nc_inq_varid(grid, args[3], &id),
	       "desired variable not found in simulation grid");

      /*
       * Line "drawing" algorithm after Bresenham
       */
      {
	int dx = x1 - idx[2];
	int dy = y1 - idx[1];
	int stepx, stepy;

        if (dx < 0) {
	  dx = -dx;
	  stepx = -1;
	}
	else
	  stepx = 1;

        if (dy < 0) {
	  dy = -dy;
	  stepy = -1;
	}
	else
	  stepy = 1;

        dx <<= 1;
        dy <<= 1;

	NC_CHECK(nc_get_var1_float(grid, id, (idx[0] == -1 ? &idx[1] : idx),
				   &v),
		 "unable to read variable value");
	printf("%s(%lu,%lu) = %g\n", args[3], idx[2], idx[1], v);
	if (arg_count == 9)
	  NC_CHECK(nc_put_var1_float(grid, id, (idx[0] == -1 ? &idx[1] : idx),
				     &nv),
		   "unable to write variable value");

        if (dx > dy) {
	  int fraction = dy - (dx >> 1);
	  while (idx[2] != x1) {
	    if (fraction >= 0) {
	      idx[1] += stepy;
	      fraction -= dx;
	    }
	    idx[2] += stepx;
	    fraction += dy;

	    NC_CHECK(nc_get_var1_float(grid, id, (idx[0] == -1 ? &idx[1] : idx),
				       &v),
		     "unable to read variable value");
	    printf("%s(%lu,%lu) = %g\n", args[3], idx[2], idx[1], v);
	    if (arg_count == 9)
	      NC_CHECK(nc_put_var1_float(grid, id, (idx[0] == -1 ? &idx[1] : idx),
					 &nv),
		       "unable to write variable value");
	  }
        }
	else {
	  int fraction = dx - (dy >> 1);
	  while (idx[1] != y1) {
	    if (fraction >= 0) {
	      idx[2] += stepx;
	      fraction -= dy;
	    }
	    idx[1] += stepy;
	    fraction += dx;

	    NC_CHECK(nc_get_var1_float(grid, id, (idx[0] == -1 ? &idx[1] : idx),
				       &v),
		     "unable to read variable value");
	    printf("%s(%lu,%lu) = %g\n", args[3], idx[2], idx[1], v);
	    if (arg_count == 9)
	      NC_CHECK(nc_put_var1_float(grid, id, (idx[0] == -1 ? &idx[1] : idx),
					 &nv),
		       "unable to write variable value");
	  }
        }
      }
    }
    EXN_ALWAYS {
      if (grid != 0) nc_close(grid);
    }
    EXN_HANDLER {
      fprintf(stderr, "%s:%d: ", exn_current.file, exn_current.line);
      switch (exn_current.code) {
      case EXN_CODE_STDLIB:
	perror(exn_current.msg);
	break;
      default:
	fputs(exn_current.msg, stderr);
	putc('\n', stderr);
      }
    }
    EXN_END
  }

  return 0;
}

/* vim:set shiftwidth=2 softtabstop=2: */
