/*!
 * @header 2D Floating Point Vectors
 * Small utilities for two dimensional floating point vectors.
 */

#ifndef __V2FLOAT__
#define __V2FLOAT__

/*!
 * @typedef v2float
 * 2D floating point vector
 * @field a the vector's two components as a floating point array
 * @field x the vector's first component
 * @field y the vector's second component
 */
typedef union v2float {
  float a[2];
  struct {
    float x, y;
  } ;
} v2float;

/*!
 * @define V2FLOAT
 * Creates a new 2D floating point vector from its two components.
 * @param x first component
 * @param y second component
 */
#define V2FLOAT(x, y) ((v2float){{(x), (y)}})

/*!
 * @define V2FLOAT0
 * The 2D floating point zero vector.
 */
#define V2FLOAT0 V2FLOAT(0, 0)

/*!
 * @define V2FLOAT_ADD
 * Adds two 2D floating point vectors.
 * @param acc first summand and storage for the result
 * @param delta second summand
 */
#define V2FLOAT_ADD(acc, delta)				\
  ((void)((acc).x += (delta).x, (acc).y += (delta).y))

/*!
 * @define V2FLOAT_SUB
 * Subtracts two 2D floating point vectors.
 * @param acc first operand and storage for the result
 * @param delta second operand
 */
#define V2FLOAT_SUB(acc, delta)				\
  ((void)((acc).x -= (delta).x, (acc).y -= (delta).y))

/*!
 * @define V2FLOAT_MUL
 * Multiplies two 2D floating point vector elementwise by each other.
 * @param acc first operand and storage for the result
 * @param mul second operand
 */
#define V2FLOAT_MUL(acc, mul)				\
  ((void)((acc).x *= (mul).x, (acc).y *= (mul).y))

/*!
 * @define V2FLOAT_DIV
 * Divides two 2D floating point vector elementwise by each other.
 * @param acc first operand and storage for the result
 * @param div second operand
 */
#define V2FLOAT_DIV(acc, div)				\
  ((void)((acc).x /= (div).x, (acc).y /= (div).y))

/*!
 * @define V2FLOAT_SMUL
 * Multiplies a 2D floating point vector by a scalar factor.
 * @param acc first operand and storage for the result
 * @param mul second operand
 */
#define V2FLOAT_SMUL(acc, mul)				\
  ((void)((acc).x *= (mul), (acc).y *= (mul)))

/*!
 * @define V2FLOAT_SDIV
 * Divides a 2D floating point vector by a scalar factor.
 * @param acc first operand and storage for the result
 * @param div second operand
 */
#define V2FLOAT_SDIV(acc, div)				\
  ((void)((acc).x /= (div), (acc).y /= (div)))

/*!
 * @define V2FLOAT_ABS
 * Returns the length of 2D floating point vector
 * @param acc vector operand
 */
#define V2FLOAT_ABS(acc)				\
  (sqrtf((acc).x * (acc).x + (acc).y * (acc).y))

#endif /* __V2FLOAT__ */
