/*!
 * @header 2D Integer Vectors
 * Small utilities for two dimensional integer vectors.
 */

#ifndef __V2INT__
#define __V2INT__

/*!
 * @typedef v2int
 * 2D integer vector
 * @field a the vector's two components as an integer array
 * @field x the vector's first component
 * @field y the vector's second component
 */
typedef union v2int {
  int a[2];
  struct {
    int x, y;
  } ;
} v2int;

/*!
 * @define V2INT
 * Creates a new 2D integer vector from its two components.
 * @param x first component
 * @param y second component
 */
#define V2INT(x, y) ((v2int){{(x), (y)}})

/*!
 * @define V2INT0
 * The 2D integer zero vector.
 */
#define V2INT0 V2INT(0, 0)

/*!
 * @define V2INT_ADD
 * Adds two 2D integer vectors.
 * @param acc first summand and storage for the result
 * @param delta second summand
 */
#define V2INT_ADD(acc, delta)				\
  ((void)((acc).x += (delta).x, (acc).y += (delta).y))

/*!
 * @define V2INT_SUB
 * Subtracts two 2D integer vectors.
 * @param acc first operand and storage for the result
 * @param delta second operand
 */
#define V2INT_SUB(acc, delta)				\
  ((void)((acc).x -= (delta).x, (acc).y -= (delta).y))

/*!
 * @define V2INT_MUL
 * Multiplies two 2D integer vector elementwise by each other.
 * @param acc first operand and storage for the result
 * @param mul second operand
 */
#define V2INT_MUL(acc, mul)				\
  ((void)((acc).x *= (mul).x, (acc).y *= (mul).y))

/*!
 * @define V2INT_DIV
 * Divides two 2D integer vector elementwise by each other.
 * @param acc first operand and storage for the result
 * @param div second operand
 */
#define V2INT_DIV(acc, div)				\
  ((void)((acc).x /= (div).x, (acc).y /= (div).y))

/*!
 * @define V2INT_MOD
 * Divides two 2D integer vector elementwise by each other and stores
 * the remainder.
 * @param acc first operand and storage for the result
 * @param div second operand
 */
#define V2INT_MOD(acc, div)				\
  ((void)((acc).x %= (div).x, (acc).y %= (div).y))

/*!
 * @define V2INT_OFS
 * Calculates an offset into a 2D array given two 2D integer vectors that
 * contain the position in the array and its dimensions. Note that the column
 * position (y coordinate) is assumed to be slow varying.
 * @param pos position in the array
 * @param dim dimensions of the array
 */
#define V2INT_OFS(pos, dim)			\
  ((pos).y * (dim).x + (pos).x)

#endif /* __V2INT__ */

/* vim:set shiftwidth=2 softtabstop=2: */
